/*
 * AIEngine a new generation network intrusion detection system.
 *
 * Copyright (C) 2013-2017  Luis Campo Giralte
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 *
 * Written by Luis Campo Giralte <luis.camp0.2009@gmail.com> 
 *
 */
#include "DNSProtocol.h"
#include <iomanip> // setw

namespace aiengine {

#ifdef HAVE_LIBLOG4CXX
log4cxx::LoggerPtr DNSProtocol::logger(log4cxx::Logger::getLogger("aiengine.dns"));
#endif

DNSProtocol::DNSProtocol():
	Protocol("DNSProtocol","dns"),
	dns_header_(nullptr),
	total_allow_queries_(0),
	total_ban_queries_(0),
	total_queries_(0),
	total_responses_(0),
	total_events_(0),
	total_dns_type_a_(0),
	total_dns_type_ns_(0),
	total_dns_type_cname_(0),
	total_dns_type_soa_(0),
	total_dns_type_ptr_(0),
	total_dns_type_mx_(0),
	total_dns_type_txt_(0),
	total_dns_type_aaaa_(0),
	total_dns_type_loc_(0),
	total_dns_type_srv_(0),
	total_dns_type_ds_(0),
	total_dns_type_dnskey_(0),
	total_dns_type_ixfr_(0),
	total_dns_type_axfr_(0),
	total_dns_type_others_(0),
	info_cache_(new Cache<DNSInfo>("DNS Info cache")),
	name_cache_(new Cache<StringCache>("Name cache")),
	domain_map_(),
	domain_mng_(),ban_domain_mng_(),
	flow_mng_(),
	current_flow_(nullptr),
        anomaly_(),
        cache_mng_() {}

DNSProtocol::~DNSProtocol() { 

	anomaly_.reset(); 
	cache_mng_.reset(); 
}

bool DNSProtocol::dnsChecker(Packet &packet) {

	// I dont like this idea of ports but...
	if (((packet.getSourcePort() == 53)||(packet.getDestinationPort() == 53))
		or((packet.getSourcePort() == 5353)||(packet.getDestinationPort() == 5353))) {
		++total_validated_packets_;
		return true;
	} else {
		++total_malformed_packets_;
		return false;
	}
}

void DNSProtocol::setDynamicAllocatedMemory(bool value) {

	info_cache_->setDynamicAllocatedMemory(value);
	name_cache_->setDynamicAllocatedMemory(value);
}

bool DNSProtocol::isDynamicAllocatedMemory() const {

	return info_cache_->isDynamicAllocatedMemory();
}

int64_t DNSProtocol::getCurrentUseMemory() const {

        int64_t mem = sizeof(DNSProtocol);

        mem += info_cache_->getCurrentUseMemory();
        mem += name_cache_->getCurrentUseMemory();

        return mem;
}

int64_t DNSProtocol::getAllocatedMemory() const {

        int64_t mem = sizeof(DNSProtocol);

        mem += name_cache_->getAllocatedMemory();
        mem += info_cache_->getAllocatedMemory();

        return mem;
}

int64_t DNSProtocol::getTotalAllocatedMemory() const {

        int64_t mem = getAllocatedMemory();

        mem += compute_memory_used_by_maps();
	
	return mem;
}

int64_t DNSProtocol::compute_memory_used_by_maps() const {

	int64_t bytes = 0;
	// Compute the size of the strings used as keys on the map
	std::for_each (domain_map_.begin(), domain_map_.end(), [&bytes] (PairStringCacheHits const &dt) {
		bytes += dt.first.size();
	});
	return bytes;
}

int32_t DNSProtocol::getTotalCacheMisses() const {

	int32_t miss = 0;

	miss = info_cache_->getTotalFails();
	miss += name_cache_->getTotalFails();

	return miss;
}


void DNSProtocol::releaseCache() {

	FlowManagerPtr fm = flow_mng_.lock();

	if (fm) {
		auto ft = fm->getFlowTable();

		std::ostringstream msg;
        	msg << "Releasing " << getName() << " cache";

		infoMessage(msg.str());

		int64_t total_bytes_released = compute_memory_used_by_maps();
		int64_t total_bytes_released_by_flows = 0;
		int32_t release_flows = 0;
		int32_t release_doms = domain_map_.size();

		for (auto &flow: ft) {
			SharedPointer<DNSInfo> info = flow->getDNSInfo();
			if (info) {
				total_bytes_released_by_flows += releaseStringToCache(name_cache_, info->name);
				total_bytes_released_by_flows += sizeof(info);
			
				++release_flows;
				flow->layer7info.reset();
				info_cache_->release(info);
			}
		} 
		domain_map_.clear();

		double cache_compression_rate = 0;

		if (total_bytes_released_by_flows > 0 ) {
			cache_compression_rate = 100 - ((total_bytes_released * 100) / total_bytes_released_by_flows);	
		}

		msg.str("");
		msg << "Release " << release_doms << " domains, " << release_flows << " flows";
		msg << ", " << total_bytes_released + total_bytes_released_by_flows << " bytes";
		msg << ", compression rate " << cache_compression_rate << "%";	
		infoMessage(msg.str());
	}
}

void DNSProtocol::attach_dns_to_flow(DNSInfo *info, boost::string_ref &domain, uint16_t qtype) {

	SharedPointer<StringCache> name = info->name;

        if (!name) { // There is no DNS attached
		GenericMapType::iterator it = domain_map_.find(domain);
                if (it == domain_map_.end()) {
                       	name = name_cache_->acquire();
                        if (name) {
                        	name->setName(domain.data(), domain.length());
				info->setQueryType(qtype);
				
                                info->name = name;
                                domain_map_.insert(std::make_pair(boost::string_ref(name->getName()),
					std::make_pair(name, 1)));
                        }
		} else {
			int *counter = &std::get<1>(it->second);
                        ++(*counter);
			info->name = std::get<0>(it->second);
		}
	}
}

void DNSProtocol::processFlow(Flow *flow) {

	int length = flow->packet->getLength();
	total_bytes_ += length;
	++total_packets_;

	current_flow_ = flow;

	if (length > header_size) { // Minimum header size consider
		setHeader(flow->packet->getPayload());

        	SharedPointer<DNSInfo> info = flow->getDNSInfo();
        	if (!info) {
                	info = info_cache_->acquire();
                	if (!info) {
#ifdef HAVE_LIBLOG4CXX
				LOG4CXX_WARN (logger, "No memory on '" << info_cache_->getName() << "' for flow:" << *flow);
#endif
				return;
                	}
                	flow->layer7info = info;
        	}

        	if (info->getIsBanned() == true) {
                	return;
        	}

		if (dns_header_->qr == 0) { // Query
			++total_queries_;
			if (ntohs(dns_header_->questions) > 0) {
				handle_standard_query(info.get(), length);
			}
		} else { // Responses
			++total_responses_; 
			if (ntohs(dns_header_->answers) > 0) {
				handle_standard_response(info.get(), length);
			}
		}
	} else {
		++total_events_;
               	if (flow->getPacketAnomaly() == PacketAnomalyType::NONE) {
               		flow->setPacketAnomaly(PacketAnomalyType::DNS_BOGUS_HEADER);
		}
                anomaly_->incAnomaly(PacketAnomalyType::DNS_BOGUS_HEADER);
	}
	return;
} 

int DNSProtocol::parse_query_name(Flow *flow) {

	int offset = extract_domain_name(&dns_header_->data[0], MAX_DNS_BUFFER_NAME);

	if (offset >= MAX_DNS_BUFFER_NAME) {
		++total_events_;
        	if (flow->getPacketAnomaly() == PacketAnomalyType::NONE) {
                	flow->setPacketAnomaly(PacketAnomalyType::DNS_LONG_NAME);
                }
                anomaly_->incAnomaly(PacketAnomalyType::DNS_LONG_NAME);
	}
	return offset;
}

// INFO: http://www.tcpipguide.com/free/t_DNSNameNotationandMessageCompressionTechnique.htm
int DNSProtocol::extract_domain_name(unsigned char *ptr, int length) {
        int offset = 1;
        short next = ptr[0];

	std::memset(&dns_buffer_name_[0], 0, MAX_DNS_BUFFER_NAME);

	while ((offset < length) and (next > 0)) {
		if (next + offset < length) {
			std::memcpy(&dns_buffer_name_[offset - 1], &ptr[offset], next);
			offset += next + 1;
			next = ptr[offset - 1];
			if (next > 0 ) {
				dns_buffer_name_[offset - 2] = '.';
			}
		} else { // There is buffer for copy but the name is too long
			int left = length - offset;
			if (next + offset < MAX_DNS_BUFFER_NAME) // copy the rest if there is space
				left = next;	

			std::memcpy(&dns_buffer_name_[offset - 1], &ptr[offset], left);
			offset = MAX_DNS_BUFFER_NAME + 1;
			break;
		}
	}

	if (offset > 1)
	    	-- offset;

	return offset;
}

void DNSProtocol::parse_response_answer(DNSInfo *info, unsigned char *ptr,int answers) {

        // Extract the IP addresses or CNAME and store on the DNSDomain just when the domain have been matched
        for (int i = 0; i < answers; ++i) {
                struct dns_address *ans = reinterpret_cast <struct dns_address*> (ptr);
                uint16_t block_length = ntohs(ans->length);
                uint16_t type = ntohs(ans->type);
                uint16_t class_type = ntohs(ans->class_type);

                if (class_type == 0x0001) { // class IN 
                        if ((type == 0x0001)and(block_length == 4)) { // IPv4
                                uint32_t ipv4addr =  ((ans->data[3] << 24) + (ans->data[2] << 16) + (ans->data[1] << 8) + ans->data[0]);
                                in_addr a;

                                a.s_addr = ipv4addr;
                                info->addIPAddress(inet_ntoa(a));
                        } else if ((type == 0x001c)and(block_length == 16)) { // IPv6
                                char ipv6addr[INET6_ADDRSTRLEN];
                                in6_addr *in6addr = (in6_addr*)&(ans->data[0]);

                                inet_ntop(AF_INET6, in6addr, ipv6addr, INET6_ADDRSTRLEN);

                                info->addIPAddress(ipv6addr);
                        } else if (type == 0x0005) { // CNAME
                                int value __attribute__((unused)) = extract_domain_name(&ans->data[0], block_length - 2);

                                info->addName(&dns_buffer_name_[0]);
                        }
                }
                // TODO: Check offset size lengths and possible anomalies       
                ptr = &(ans->data[block_length]);
        }
}

void DNSProtocol::handle_standard_query(DNSInfo *info, int length) {
	boost::string_ref domain;
	int offset = parse_query_name(current_flow_); 
	
	boost::string_ref dns_name(dns_buffer_name_, offset);

	if (offset == 1) { // There is no name, a root record
		offset = 0;
		domain = "<Root>";
	} else {
		domain = dns_name.substr(0, offset - 1);
	}

	// Check if the payload is malformed
	if (header_size + offset > length) {
		++total_events_;
               	if (current_flow_->getPacketAnomaly() == PacketAnomalyType::NONE) {
               		current_flow_->setPacketAnomaly(PacketAnomalyType::DNS_BOGUS_HEADER);
		}
               	anomaly_->incAnomaly(current_flow_, PacketAnomalyType::DNS_BOGUS_HEADER);
		return;
	}

	uint16_t qtype = ntohs((dns_header_->data[offset + 2] << 8) + dns_header_->data[offset + 1]);

	update_query_types(qtype);

	if (domain.length() > 0) { // The domain is valid
		if (ban_domain_mng_) {
			auto domain_candidate = ban_domain_mng_->getDomainName(domain);
			if (domain_candidate) {
#ifdef HAVE_LIBLOG4CXX
				LOG4CXX_INFO (logger, "Flow:" << *current_flow_ << " matchs with banned domain " << domain_candidate->getName());
#endif
				info->setIsBanned(true);	
				++total_ban_queries_;
				return;
			}
		}

		++total_allow_queries_;
		
		attach_dns_to_flow(info,domain, qtype);	
	}
}

void DNSProtocol::handle_standard_response(DNSInfo *info, int length) {
	boost::string_ref domain;

       	int offset = parse_query_name(current_flow_);

       	SharedPointer<StringCache> name = info->name;
       	if (!name) {
		// There is no name attached so lets try to extract from the response
        	// int offset = parse_query_name(current_flow_);

        	boost::string_ref dns_name(dns_buffer_name_, offset);

        	if (offset == 1) { // There is no name, a root record
                	offset = 0;
                	domain = "<Root>";
        	} else {
                	domain = dns_name.substr(0, offset - 1);
        	}

	        uint16_t qtype = ntohs((dns_header_->data[offset + 2] << 8) + dns_header_->data[offset + 1]);

        	update_query_types(qtype);

		attach_dns_to_flow(info, domain, qtype);
	} else {
		domain = name->getName();
	}

	// Check if the DNSProtocol have a DomainNameManager attached for match domains
        if (domain_mng_) {
        	auto domain_candidate = domain_mng_->getDomainName(domain);
                if (domain_candidate) {
			++total_events_;
			// Need to increase by 4 the generate offset due to the type and class dns fields
			offset = offset + 5;
			uint16_t answers = ntohs(dns_header_->answers);
			u_char *ptr = &(dns_header_->data[offset]);

#ifdef HAVE_LIBLOG4CXX
                        LOG4CXX_INFO (logger, "Flow:" << *current_flow_ << " matchs with " << domain_candidate->getName());
#endif
			parse_response_answer(info,ptr,answers);

			info->matched_domain_name = domain_candidate;
#if defined(PYTHON_BINDING) || defined(RUBY_BINDING) || defined(JAVA_BINGING) || defined(LUA_BINDING)
			current_flow_->packet->setForceAdaptorWrite(true); // The udp layer will call the databaseAdaptor update method
                        if (domain_candidate->call.haveCallback()) {
                                domain_candidate->call.executeCallback(current_flow_);
                        }
#endif
                }
	}
}

void DNSProtocol::update_query_types(uint16_t type) {

	if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_A))
		++ total_dns_type_a_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_NS))
		++ total_dns_type_ns_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_CNAME))
		++ total_dns_type_cname_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_SOA))
		++ total_dns_type_soa_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_PTR))
		++ total_dns_type_ptr_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_MX))
		++ total_dns_type_mx_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_TXT))
		++ total_dns_type_txt_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_AAAA))
		++ total_dns_type_aaaa_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_LOC))
		++ total_dns_type_loc_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_SRV))
		++ total_dns_type_srv_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_DS))
		++ total_dns_type_ds_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_DNSKEY))
		++ total_dns_type_dnskey_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_IXFR))
		++ total_dns_type_ixfr_;
	else if (type == static_cast<uint16_t>(DNSQueryTypes::DNS_TYPE_AXFR))
		++ total_dns_type_axfr_;
	else
		++ total_dns_type_others_;

}

void DNSProtocol::setDomainNameManager(const SharedPointer<DomainNameManager>& dnm) {

	if (domain_mng_) {
		domain_mng_->setPluggedToName("");
	}
	if (dnm) {
        	domain_mng_ = dnm;
	        domain_mng_->setPluggedToName(getName());
	} else {
		domain_mng_.reset();
	}
}

void DNSProtocol::statistics(std::basic_ostream<char>& out, int level) {

	if (level > 0) {
                int64_t alloc_memory = getAllocatedMemory();
                std::string unit = "Bytes";
		const char *dynamic_memory = (isDynamicAllocatedMemory() ? "yes":"no");

                unitConverter(alloc_memory, unit);

                out << getName() << "(" << this <<") statistics" << std::dec << std::endl;

		if (ban_domain_mng_) out << "\t" << "Plugged banned domains from:" << ban_domain_mng_->getName() << std::endl;
		if (domain_mng_) out << "\t" << "Plugged domains from:" << domain_mng_->getName() << std::endl;

		out << "\t" << "Dynamic memory alloc:   " << std::setw(10) << dynamic_memory << std::endl;
                out << "\t" << "Total allocated:        " << std::setw(9 - unit.length()) << alloc_memory << " " << unit <<std::endl;
        	out << "\t" << "Total packets:          " << std::setw(10) << total_packets_ <<std::endl;
        	out << "\t" << "Total bytes:            " << std::setw(10) << total_bytes_ <<std::endl;
		if (level > 1) {
		
			out << "\t" << "Total validated packets:" << std::setw(10) << total_validated_packets_ <<std::endl;
			out << "\t" << "Total malformed packets:" << std::setw(10) << total_malformed_packets_ <<std::endl;
			if (level > 3) {
			
				out << "\t" << "Total allow queries:    " << std::setw(10) << total_allow_queries_ <<std::endl;
				out << "\t" << "Total banned queries:   " << std::setw(10) << total_ban_queries_ <<std::endl;
				out << "\t" << "Total queries:          " << std::setw(10) << total_queries_ <<std::endl;
				out << "\t" << "Total responses:        " << std::setw(10) << total_responses_ <<std::endl;
				out << "\t" << "Total type A:           " << std::setw(10) << total_dns_type_a_ <<std::endl;
				out << "\t" << "Total type NS:          " << std::setw(10) << total_dns_type_ns_ <<std::endl;
				out << "\t" << "Total type CNAME:       " << std::setw(10) << total_dns_type_cname_ <<std::endl;
				out << "\t" << "Total type SOA:         " << std::setw(10) << total_dns_type_soa_ <<std::endl;
				out << "\t" << "Total type PTR:         " << std::setw(10) << total_dns_type_ptr_ <<std::endl;
				out << "\t" << "Total type MX:          " << std::setw(10) << total_dns_type_mx_ <<std::endl;
				out << "\t" << "Total type TXT:         " << std::setw(10) << total_dns_type_txt_ <<std::endl;
				out << "\t" << "Total type AAAA:        " << std::setw(10) << total_dns_type_aaaa_ <<std::endl;
				out << "\t" << "Total type LOC:         " << std::setw(10) << total_dns_type_loc_ <<std::endl;
				out << "\t" << "Total type SRV:         " << std::setw(10) << total_dns_type_srv_ <<std::endl;
				out << "\t" << "Total type DS:          " << std::setw(10) << total_dns_type_ds_ <<std::endl;
				out << "\t" << "Total type DNSKEY:      " << std::setw(10) << total_dns_type_dnskey_ <<std::endl;
				out << "\t" << "Total type IXFR:        " << std::setw(10) << total_dns_type_ixfr_ <<std::endl;
				out << "\t" << "Total type AXFR:        " << std::setw(10) << total_dns_type_axfr_ <<std::endl;
				out << "\t" << "Total type others:      " << std::setw(10) << total_dns_type_others_ <<std::endl;
			}
			if (level > 2) {	
			
				if (flow_forwarder_.lock())
					flow_forwarder_.lock()->statistics(out);
                                if (level > 3) {
                               
					info_cache_->statistics(out); 
                                        name_cache_->statistics(out);
                                        if (level > 4) {
                                              	showCacheMap(out,domain_map_, "DNS Name", "Domain"); 
                                        }
                                }
			}
		}
	}
}


void DNSProtocol::increaseAllocatedMemory(int value) { 

	info_cache_->create(value);
	name_cache_->create(value);
}

void DNSProtocol::decreaseAllocatedMemory(int value) { 

	info_cache_->destroy(value);
	name_cache_->destroy(value);
}

CounterMap DNSProtocol::getCounters() const {
	CounterMap cm;

        cm.addKeyValue("total allow queries", total_allow_queries_);
        cm.addKeyValue("total banned queries", total_ban_queries_);
        cm.addKeyValue("total queries", total_queries_);
        cm.addKeyValue("total responses", total_responses_);
        cm.addKeyValue("total type A", total_dns_type_a_);
        cm.addKeyValue("total type NS", total_dns_type_ns_);
        cm.addKeyValue("total type CNAME", total_dns_type_cname_);
        cm.addKeyValue("total type SOA", total_dns_type_soa_);
        cm.addKeyValue("total type PTR", total_dns_type_ptr_);
        cm.addKeyValue("total type MX", total_dns_type_mx_);
        cm.addKeyValue("total type TXT", total_dns_type_txt_);
        cm.addKeyValue("total type AAAA", total_dns_type_aaaa_);
        cm.addKeyValue("total type LOC", total_dns_type_loc_);
        cm.addKeyValue("total type SRV", total_dns_type_srv_);
        cm.addKeyValue("total type DS", total_dns_type_ds_);
        cm.addKeyValue("total type DNSKEY", total_dns_type_dnskey_);
        cm.addKeyValue("total type IXFR", total_dns_type_ixfr_);
        cm.addKeyValue("total type AXFR", total_dns_type_axfr_);
        cm.addKeyValue("total type others", total_dns_type_others_);

        return cm;
}

#if defined(PYTHON_BINDING) || defined(RUBY_BINDING)
#if defined(PYTHON_BINDING)
boost::python::dict DNSProtocol::getCache() const {
#elif defined(RUBY_BINDING)
VALUE DNSProtocol::getCache() const {
#endif
	return addMapToHash(domain_map_);
}
#endif

} // namespace aiengine

