/*
 * AIEngine a new generation network intrusion detection system.
 *
 * Copyright (C) 2013-2017  Luis Campo Giralte
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 *
 * Written by Luis Campo Giralte <luis.camp0.2009@gmail.com> 
 *
 */
#include "SIPProtocol.h"
#include <iomanip> // setw

namespace aiengine {

#ifdef HAVE_LIBLOG4CXX
log4cxx::LoggerPtr SIPProtocol::logger(log4cxx::Logger::getLogger("aiengine.sip"));
#endif

// List of support request methods 
std::vector<SipMethodType> SIPProtocol::methods_ {
	std::make_tuple("REGISTER"	,8,	"registers"	,0),
	std::make_tuple("INVITE"	,6,	"invites"	,0),
	std::make_tuple("ACK"		,3,	"acks"		,0),
	std::make_tuple("CANCEL"	,6,	"cancels"	,0),
	std::make_tuple("BYE"		,3,	"byes"		,0),
	std::make_tuple("MESSAGE"	,7,	"messages"	,0),
	std::make_tuple("OPTIONS"	,7,	"options"	,0),
	std::make_tuple("PUBLISH"	,7,	"publishs"	,0),
	std::make_tuple("SUBSCRIBE"	,9,	"subcribes"	,0),
	std::make_tuple("NOTIFY"	,6,	"notifies"	,0),
	std::make_tuple("REFER"		,5,	"refers"	,0),
	std::make_tuple("INFO"		,4,	"infos"		,0),
	std::make_tuple("PING"		,4,	"pings"		,0)
};

SIPProtocol::SIPProtocol():
	Protocol("SIPProtocol","sip"),
	sip_regex_(new Regex("Main SIP expression","^(REGISTER|INVITE|OPTIONS).*SIP/2.")),
	sip_from_(new Regex("From expression","From: .*?\r\n")),
	sip_to_(new Regex("To expression","To: .*?\r\n")),
	sip_via_(new Regex("Via expression","Via: .*?\r\n")),
	sip_header_(nullptr),
	total_events_(0),
	total_requests_(0),
	total_responses_(0),
	total_sip_others_(0),
	info_cache_(new Cache<SIPInfo>("SIP Info cache")),
	uri_cache_(new Cache<StringCache>("Uri cache")),
	via_cache_(new Cache<StringCache>("Via cache")),
	from_cache_(new Cache<StringCache>("From cache")),
	to_cache_(new Cache<StringCache>("To cache")),
	uri_map_(),
	via_map_(),
	from_map_(),
	to_map_(),
	flow_mng_(),
	cache_mng_() 
	{}

SIPProtocol::~SIPProtocol() { 

	cache_mng_.reset(); 
}

bool SIPProtocol::sipChecker(Packet& packet) {

	// TODO: I dont like this idea of ports but...
	if ((packet.getSourcePort() == 5060)||(packet.getDestinationPort() == 5060)) {

		setHeader(packet.getPayload());
		++total_validated_packets_;
		return true;
	} else {
		++total_malformed_packets_;
		return false;
	}
}

void SIPProtocol::setDynamicAllocatedMemory(bool value) {

	info_cache_->setDynamicAllocatedMemory(value);
	uri_cache_->setDynamicAllocatedMemory(value);
	via_cache_->setDynamicAllocatedMemory(value);
	from_cache_->setDynamicAllocatedMemory(value);
	to_cache_->setDynamicAllocatedMemory(value);
}

bool SIPProtocol::isDynamicAllocatedMemory() const {

	return info_cache_->isDynamicAllocatedMemory();
}	

int64_t SIPProtocol::getCurrentUseMemory() const {

	int64_t mem = sizeof(SIPProtocol);

	mem += info_cache_->getCurrentUseMemory();
	mem += uri_cache_->getCurrentUseMemory();
	mem += via_cache_->getCurrentUseMemory();
	mem += from_cache_->getCurrentUseMemory();
	mem += to_cache_->getCurrentUseMemory();
	
	return mem;
}

int64_t SIPProtocol::getAllocatedMemory() const {

        int64_t mem = sizeof(SIPProtocol);

        mem += info_cache_->getAllocatedMemory();
        mem += uri_cache_->getAllocatedMemory();
        mem += via_cache_->getAllocatedMemory();
        mem += from_cache_->getAllocatedMemory();
        mem += to_cache_->getAllocatedMemory();

        return mem;
}

int32_t SIPProtocol::release_sip_info(SIPInfo *info) {

        int32_t bytes_released = 0;

        bytes_released = releaseStringToCache(uri_cache_, info->uri);
        bytes_released += releaseStringToCache(via_cache_, info->via);
        bytes_released += releaseStringToCache(from_cache_, info->from);
        bytes_released += releaseStringToCache(to_cache_, info->to);

        return bytes_released;
}

int64_t SIPProtocol::getTotalAllocatedMemory() const {

        int64_t mem = getAllocatedMemory();
	
	mem += compute_memory_used_by_maps();

	return mem;
}

int64_t SIPProtocol::compute_memory_used_by_maps() const {

	int64_t bytes = 0;
	
	// Compute the size of the strings used as keys on the map
	std::for_each (from_map_.begin(), from_map_.end(), [&bytes] (PairStringCacheHits const &f) {
		bytes += f.first.size();
	});
	std::for_each (uri_map_.begin(), uri_map_.end(), [&bytes] (PairStringCacheHits const &u) {
		bytes += u.first.size();
	});
	std::for_each (to_map_.begin(), to_map_.end(), [&bytes] (PairStringCacheHits const &t) {
		bytes += t.first.size();
	});
	std::for_each (via_map_.begin(), via_map_.end(), [&bytes] (PairStringCacheHits const &t) {
		bytes += t.first.size();
	});
	return bytes;
}

int32_t SIPProtocol::getTotalCacheMisses() const {

	int32_t miss = 0;

	miss = info_cache_->getTotalFails();
	miss += uri_cache_->getTotalFails();
	miss += via_cache_->getTotalFails();
	miss += from_cache_->getTotalFails();
	miss += to_cache_->getTotalFails();

	return miss;
}

void SIPProtocol::releaseCache() {

        FlowManagerPtr fm = flow_mng_.lock();

        if (fm) {
                auto ft = fm->getFlowTable();

                std::ostringstream msg;
                msg << "Releasing " << getName() << " cache";

                infoMessage(msg.str());

                int64_t total_bytes_released = compute_memory_used_by_maps();
                int64_t total_bytes_released_by_flows = 0;
                int32_t release_flows = 0;
                int32_t release_from = from_map_.size();
                int32_t release_uris = uri_map_.size();
                int32_t release_to = to_map_.size();
                int32_t release_via = via_map_.size();

                for (auto &flow: ft) {
			SharedPointer<SIPInfo> info = flow->getSIPInfo();
			if (info) {
				total_bytes_released_by_flows += release_sip_info(info.get());
                		total_bytes_released_by_flows += sizeof(info);        
	
                        	++release_flows;
				flow->layer7info.reset();
				info_cache_->release(info);
			}
                } 
                uri_map_.clear();
                from_map_.clear();
                to_map_.clear();
                via_map_.clear();

                double cache_compression_rate = 0;

                if (total_bytes_released_by_flows > 0 ) {
                        cache_compression_rate = 100 - ((total_bytes_released * 100) / total_bytes_released_by_flows);
                }
        
        	msg.str("");
                msg << "Release " << release_uris << " uris, " << release_via << " vias, " << release_from;
                msg << " froms, " << release_to << " tos, " << release_flows << " flows";
                msg << ", " << total_bytes_released << " bytes, compression rate " << cache_compression_rate << "%";
                infoMessage(msg.str());
        }
}

void SIPProtocol::extract_via_value(SIPInfo *info, boost::string_ref &header) {

        if (sip_via_->matchAndExtract(header)) {

		boost::string_ref via_raw(sip_via_->getExtract());
		boost::string_ref via(via_raw.substr(5, via_raw.size() - 7)); // remove also the \r\n
                
                attach_via_to_flow(info, via);
        }
}

void SIPProtocol::extract_from_value(SIPInfo *info, boost::string_ref &header) {

	if (sip_from_->matchAndExtract(header)) {

		boost::string_ref from_raw(sip_from_->getExtract());
		boost::string_ref from(from_raw.substr(6, from_raw.size() - 8)); // remove also the \r\n
 
		attach_from_to_flow(info, from);
	}
}


void SIPProtocol::attach_from_to_flow(SIPInfo *info, boost::string_ref &from) {

	if (!info->from) {
		GenericMapType::iterator it = from_map_.find(from);
		if (it == from_map_.end()) {
			SharedPointer<StringCache> from_ptr = from_cache_->acquire();
			if (from_ptr) {
				from_ptr->setName(from.data(), from.length());
				info->from = from_ptr;
				from_map_.insert(std::make_pair(boost::string_ref(from_ptr->getName()),
					std::make_pair(from_ptr, 1)));
			}
		} else {
			int *counter = &std::get<1>(it->second);
			++(*counter);
			info->from = std::get<0>(it->second);
		}
	}
}

void SIPProtocol::extract_to_value(SIPInfo *info, boost::string_ref &header) {

	if (sip_to_->matchAndExtract(header)) {

		boost::string_ref to_raw(sip_to_->getExtract());
		boost::string_ref to(to_raw.substr(4, to_raw.size() - 6)); // remove also the \r\n

		attach_to_to_flow(info, to);
	}
}

void SIPProtocol::attach_to_to_flow(SIPInfo *info, boost::string_ref &to) {

	if (!info->to) {
		GenericMapType::iterator it = to_map_.find(to);
		if (it == to_map_.end()) {
			SharedPointer<StringCache> to_ptr = to_cache_->acquire();
			if (to_ptr) {
				to_ptr->setName(to.data(), to.length());
				info->to = to_ptr;
				to_map_.insert(std::make_pair(boost::string_ref(to_ptr->getName()),
					std::make_pair(to_ptr, 1)));
			}	
		} else {
			int *counter = &std::get<1>(it->second);
			++(*counter);
			info->to = std::get<0>(it->second);	
		}
	}

}

void SIPProtocol::attach_via_to_flow(SIPInfo *info, boost::string_ref &via) {

	if (!info->via) {
                GenericMapType::iterator it = via_map_.find(via);
                if (it == via_map_.end()) {
                        SharedPointer<StringCache> via_ptr = via_cache_->acquire();
                        if (via_ptr) {
                                via_ptr->setName(via.data(), via.length());
                                info->via = via_ptr;
                                via_map_.insert(std::make_pair(boost::string_ref(via_ptr->getName()),
					std::make_pair(via_ptr, 1)));
                        }
                } else {
                        int *counter = &std::get<1>(it->second);
                        ++(*counter);
                        info->via = std::get<0>(it->second);
                }
        }
}


void SIPProtocol::attach_uri_to_flow(SIPInfo *info, boost::string_ref &uri) {

	GenericMapType::iterator it = uri_map_.find(uri);
        if (it == uri_map_.end()) {
        	SharedPointer<StringCache> uri_ptr = uri_cache_->acquire();
                if (uri_ptr) {
                	uri_ptr->setName(uri.data(), uri.length());
                        info->uri = uri_ptr;
                        uri_map_.insert(std::make_pair(boost::string_ref(uri_ptr->getName()),
				std::make_pair(uri_ptr, 1)));
                } 
        } else {
		// Update the URI of the flow
                int *counter = &std::get<1>(it->second);
                ++(*counter);
		info->uri = std::get<0>(it->second);
	}
}


void SIPProtocol::extract_uri_value(SIPInfo *info, boost::string_ref &header) {

	int offset = 0;
	bool found = false;

	// Check if is a response 
        if (std::memcmp("SIP/2.", &header[0], 6) == 0) {
                ++total_responses_;

                // No uri to extract
		return;
	} 

	for (auto &method: methods_) {
		const char *m = std::get<0>(method);
		offset = std::get<1>(method);

		if (std::memcmp(m, &header[0], offset) == 0) {
			int32_t *hits = &std::get<3>(method);

			found = true;
			++offset;
			++(*hits);
			break;
		}
	}

	if ((found)and(offset > 0)) {
		int end = header.find("SIP/2.");
		if (end > 0) {
			boost::string_ref uri(header.substr(offset, (end - offset) - 1));
	
			++total_requests_;	
			attach_uri_to_flow(info, uri);	
		}
	}else{
		++total_sip_others_;
	}
}

void SIPProtocol::processFlow(Flow *flow) {

	++total_packets_;	
	int length = flow->packet->getLength();
	total_bytes_ += length;
	++flow->total_packets_l7;

	SharedPointer<SIPInfo> info = flow->getSIPInfo();

        if (!info) {
                info = info_cache_->acquire();
                if (!info) {
#ifdef HAVE_LIBLOG4CXX
			LOG4CXX_WARN (logger, "No memory on '" << info_cache_->getName() << "' for flow:" << *flow);
#endif
			return;
                }
                flow->layer7info = info;
        }

	boost::string_ref header(reinterpret_cast <const char*> (flow->packet->getPayload()), length);

	extract_uri_value(info.get(),header);
	
	extract_via_value(info.get(),header);
		
	extract_from_value(info.get(),header);	

	extract_to_value(info.get(),header);
}


void SIPProtocol::increaseAllocatedMemory(int value) {

	info_cache_->create(value);
	uri_cache_->create(value);
	from_cache_->create(value);
	to_cache_->create(value);
	via_cache_->create(value);
}


void SIPProtocol::decreaseAllocatedMemory(int value) {

	info_cache_->destroy(value);
	uri_cache_->destroy(value);
	from_cache_->destroy(value);
	to_cache_->destroy(value);
	via_cache_->destroy(value);
}

void SIPProtocol::statistics(std::basic_ostream<char>& out, int level) {

	if (level > 0) {
                int64_t alloc_memory = getAllocatedMemory();
                std::string unit = "Bytes";
		const char *dynamic_memory = (isDynamicAllocatedMemory() ? "yes":"no");

                unitConverter(alloc_memory, unit);

                out << getName() << "(" << this <<") statistics" << std::dec << std::endl;
		out << "\t" << "Dynamic memory alloc:   " << std::setw(10) << dynamic_memory << std::endl;
                out << "\t" << "Total allocated:        " << std::setw(9 - unit.length()) << alloc_memory << " " << unit <<std::endl;
		out << "\t" << "Total packets:          " << std::setw(10) << total_packets_ <<std::endl;
		out << "\t" << "Total bytes:        " << std::setw(14) << total_bytes_ <<std::endl;
		if (level > 1) {
        		out << "\t" << "Total validated packets:" << std::setw(10) << total_validated_packets_ <<std::endl;
        		out << "\t" << "Total malformed packets:" << std::setw(10) << total_malformed_packets_ <<std::endl;
			if (level > 3) { 
			
				out << "\t" << "Total requests:         " << std::setw(10) << total_requests_ <<std::endl;
				out << "\t" << "Total responses:        " << std::setw(10) << total_responses_ <<std::endl;
				for (auto &method: methods_) {
					const char *label = std::get<2>(method);
					int32_t hits = std::get<3>(method);
					out << "\t" << "Total " << label << ":" << std::right << std::setfill(' ') << std::setw(27 - strlen(label)) << hits <<std::endl;

				}
				out << "\t" << "Total others:           " << std::setw(10) << total_sip_others_ <<std::endl;
			}
			if (level > 2) {
				if (flow_forwarder_.lock())
					flow_forwarder_.lock()->statistics(out);
				if (level > 3) {
					info_cache_->statistics(out);
					uri_cache_->statistics(out);
					via_cache_->statistics(out);
					from_cache_->statistics(out);
					to_cache_->statistics(out);
					if (level > 4) {
						showCacheMap(out,uri_map_, "SIP Uris", "Uri");
						showCacheMap(out,via_map_, "SIP Vias", "Via");
						showCacheMap(out,from_map_, "SIP Froms", "From");
						showCacheMap(out,to_map_, "SIP Tos", "To");
					}
				}
			}
		}
	}
}

CounterMap SIPProtocol::getCounters() const {
        CounterMap cm;

        cm.addKeyValue("packets", total_packets_);
        cm.addKeyValue("bytes", total_bytes_);
        cm.addKeyValue("requests", total_requests_);
        cm.addKeyValue("responses", total_requests_);
	for (auto &method: methods_) {
		const char *label = std::get<2>(method);

		cm.addKeyValue(label, std::get<3>(method));
	}
        cm.addKeyValue("others", total_sip_others_);
        return cm;
}

} // namespace aiengine 
